@@ -38,9 +38,9 @@ from The Open Group.
 #ifdef DPMSExtension
 #include "dpmsproc.h"
 #endif
-#include <locale.h>
 #ifdef __CYGWIN__
 #include <mntent.h>
+#include <sys/statvfs.h>
 #endif
 #if defined(WIN32)
 #include "xkbsrv.h"
@@ -66,6 +66,7 @@ typedef WINAPI HRESULT(*SHGETFOLDERPATHPROC) (HWND hwndOwner,
 #include "glx/glwindows.h"
 #include "dri/windowsdri.h"
 #endif
+#include "winauth.h"
 
 /*
  * References to external symbols
@@ -90,9 +91,6 @@ const char *winGetBaseDir(void);
 
 static void winCheckMount(void);
 
-extern Bool XSupportsLocale(void);
-extern Status XInitThreads(void);
-
 /*
  * For the depth 24 pixmap we default to 32 bits per pixel, but
  * we change this pixmap format later if we detect that the display
@@ -116,15 +114,10 @@ static PixmapFormatRec g_PixmapFormats[] = {
     {32, 32, BITMAP_SCANLINE_PAD}
 };
 
-const int NUMFORMATS = sizeof(g_PixmapFormats) / sizeof(g_PixmapFormats[0]);
-
 static const ExtensionModule xwinExtensions[] = {
-#ifdef GLXEXT
-  { GlxExtensionInit, "GLX", &noGlxExtension },
 #ifdef XWIN_WINDOWS_DRI
   { WindowsDRIExtensionInit, "Windows-DRI", &noDriExtension },
 #endif
-#endif
 };
 
 /*
@@ -138,10 +131,11 @@ void XwinExtensionInit(void)
     if (g_fNativeGl) {
         /* install the native GL provider */
         glxWinPushNativeProvider();
+
+        /* load the Windows-DRI extension */
+        LoadExtensionList(xwinExtensions, ARRAY_SIZE(xwinExtensions), TRUE);
     }
 #endif
-
-    LoadExtensionList(xwinExtensions, ARRAY_SIZE(xwinExtensions), TRUE);
 }
 
 #if defined(DDXBEFORERESET)
@@ -155,9 +149,7 @@ ddxBeforeReset(void)
 {
     winDebug("ddxBeforeReset - Hello\n");
 
-#ifdef XWIN_CLIPBOARD
     winClipboardShutdown();
-#endif
 }
 #endif
 
@@ -201,13 +193,11 @@ ddxGiveUp(enum ExitCode error)
             winDeleteNotifyIcon(winGetScreenPriv(g_ScreenInfo[i].pScreen));
     }
 
-#ifdef XWIN_MULTIWINDOW
     /* Unload libraries for taskbar grouping */
     winPropertyStoreDestroy();
 
     /* Notify the worker threads we're exiting */
     winDeinitMultiWindowWM();
-#endif
 
 #ifdef HAS_DEVWINDOWS
     /* Close our handle to our message queue */
@@ -301,7 +291,7 @@ winCheckMount(void)
     enum { none = 0, sys_root, user_root, sys_tmp, user_tmp }
         level = none, curlevel;
     BOOL binary = TRUE;
-    BOOL fat = TRUE;
+    struct statvfs buf;
 
     mnt = setmntent("/etc/mtab", "r");
     if (mnt == NULL) {
@@ -340,12 +330,6 @@ winCheckMount(void)
             binary = FALSE;
         else
             binary = TRUE;
-
-        if ((strcmp(ent->mnt_type, "vfat") == 0) ||
-            (strcmp(ent->mnt_type, "exfat") == 0))
-            fat = TRUE;
-        else
-            fat = FALSE;
     }
 
     if (endmntent(mnt) != 1) {
@@ -356,10 +340,19 @@ winCheckMount(void)
     if (!binary)
         winMsg(X_WARNING, "/tmp mounted in textmode\n");
 
-    if (fat) {
-        winMsg(X_WARNING,
-               "/tmp mounted on FAT filesystem, activating -nolock\n");
-        nolock = TRUE;
+    /*
+       Use statvfs(), which on Cygwin passes through the flags returned by
+       GetVolumeInformation(), to determine if hardlinks are available
+
+       This should use the same LOCK_DIR as LockServer()
+    */
+    if (!statvfs("/tmp", &buf)) {
+        if (!(buf.f_flag & FILE_SUPPORTS_HARD_LINKS)) {
+            winMsg(X_WARNING, "/tmp mounted on a filesystem without hardlinks, activating -nolock\n");
+            nolock = TRUE;
+        }
+    } else {
+        winMsg(X_WARNING, "statvfs() /tmp failed\n");
     }
 }
 #else
@@ -739,16 +732,21 @@ winUseMsg(void)
     ErrorF(EXECUTABLE_NAME " Device Dependent Usage:\n");
     ErrorF("\n");
 
-#ifdef XWIN_CLIPBOARD
     ErrorF("-[no]clipboard\n"
            "\tEnable [disable] the clipboard integration. Default is enabled.\n");
-#endif
 
     ErrorF("-clipupdates num_boxes\n"
            "\tUse a clipping region to constrain shadow update blits to\n"
            "\tthe updated region when num_boxes, or more, are in the\n"
            "\tupdated region.\n");
 
+    ErrorF("-[no]compositealpha\n"
+           "\tX windows with per-pixel alpha are composited into the Windows desktop.\n");
+    ErrorF("-[no]compositewm\n"
+           "\tUse the Composite extension to keep a bitmap image of each top-level\n"
+           "\tX window, so window contents which are occluded show correctly in\n"
+           "\ttask bar and task switcher previews.\n");
+
 #ifdef XWIN_XF86CONFIG
     ErrorF("-config\n" "\tSpecify a configuration file.\n");
 
@@ -816,28 +814,21 @@ winUseMsg(void)
            "\tUse the entire virtual screen if multiple\n"
            "\tmonitors are present.\n");
 
-#ifdef XWIN_MULTIWINDOW
     ErrorF("-multiwindow\n" "\tRun the server in multi-window mode.\n");
-#endif
 
     ErrorF("-nodecoration\n"
            "\tDo not draw a window border, title bar, etc.  Windowed\n"
            "\tmode only.\n");
 
-#ifdef XWIN_CLIPBOARD
-    ErrorF("-nounicodeclipboard\n"
-           "\tDo not use Unicode clipboard even if on a NT-based platform.\n");
-
     ErrorF("-[no]primary\n"
            "\tWhen clipboard integration is enabled, map the X11 PRIMARY selection\n"
            "\tto the Windows clipboard. Default is enabled.\n");
-#endif
 
     ErrorF("-refresh rate_in_Hz\n"
            "\tSpecify an optional refresh rate to use in fullscreen mode\n"
            "\twith a DirectDraw engine.\n");
 
-    ErrorF("-resize=none|scrollbars|randr"
+    ErrorF("-resize=none|scrollbars|randr\n"
            "\tIn windowed mode, [don't] allow resizing of the window. 'scrollbars'\n"
            "\tmode gives the window scrollbars as needed, 'randr' mode uses the RANR\n"
            "\textension to resize the X screen.  'randr' is the default.\n");
@@ -950,10 +941,10 @@ InitOutput(ScreenInfo * pScreenInfo, int argc, char *argv[])
     pScreenInfo->bitmapScanlinePad = BITMAP_SCANLINE_PAD;
     pScreenInfo->bitmapScanlineUnit = BITMAP_SCANLINE_UNIT;
     pScreenInfo->bitmapBitOrder = BITMAP_BIT_ORDER;
-    pScreenInfo->numPixmapFormats = NUMFORMATS;
+    pScreenInfo->numPixmapFormats = ARRAY_SIZE(g_PixmapFormats);
 
     /* Describe how we want common pixmap formats padded */
-    for (i = 0; i < NUMFORMATS; i++) {
+    for (i = 0; i < ARRAY_SIZE(g_PixmapFormats); i++) {
         pScreenInfo->formats[i] = g_PixmapFormats[i];
     }
 
@@ -962,10 +953,8 @@ InitOutput(ScreenInfo * pScreenInfo, int argc, char *argv[])
 
     /* Detect supported engines */
     winDetectSupportedEngines();
-#ifdef XWIN_MULTIWINDOW
     /* Load libraries for taskbar grouping */
     winPropertyStoreInit();
-#endif
 
     /* Store the instance handle */
     g_hInstance = GetModuleHandle(NULL);
@@ -1003,8 +992,7 @@ InitOutput(ScreenInfo * pScreenInfo, int argc, char *argv[])
           for (iMonitor = 1; ; iMonitor++)
             {
               struct GetMonitorInfoData data;
-              QueryMonitor(iMonitor, &data);
-              if (data.bMonitorSpecifiedExists)
+              if (QueryMonitor(iMonitor, &data))
                 {
                   MONITORINFO mi;
                   mi.cbSize = sizeof(MONITORINFO);
@@ -1035,37 +1023,12 @@ InitOutput(ScreenInfo * pScreenInfo, int argc, char *argv[])
         }
     }
 
-#if defined(XWIN_CLIPBOARD) || defined(XWIN_MULTIWINDOW)
+    xorgGlxCreateVendor();
 
     /* Generate a cookie used by internal clients for authorization */
     if (g_fXdmcpEnabled || g_fAuthEnabled)
         winGenerateAuthorization();
 
-    /* Perform some one time initialization */
-    if (1 == serverGeneration) {
-        const char *locale;
-
-        /* Allow multiple threads to access Xlib */
-        if (XInitThreads() == 0) {
-            ErrorF("XInitThreads failed.\n");
-        }
-
-        /*
-         * setlocale applies to all threads in the current process.
-         * Apply locale specified in LANG environment variable.
-         */
-        locale = setlocale(LC_ALL, "");
-        if (!locale) {
-            ErrorF("setlocale failed.\n");
-        }
-
-        /* See if X supports the current locale */
-        if (XSupportsLocale() == FALSE) {
-            ErrorF("Warning: Locale '%s' not supported by X, falling back to 'C' locale.\n", locale);
-            setlocale(LC_ALL, "C");
-        }
-    }
-#endif
 
 #if CYGDEBUG || YES
     winDebug("InitOutput - Returning.\n");
